--[[
Copyright (c) 2015 - 2016 Robert Machmer

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the
Software without restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
]]
local dir	= 'Graphoon'
local Node	= require( dir .. '.Node' )
local gr, lfs, floor = love.graphics, love.filesystem, math.floor
gr.setLineWidth(2)
--------------------------
-- local support functions
--------------------------
local function split( str, sep )	-- inserts delimited string values into table
	local sep = sep or "%s"			-- default delimiter = space
	local t = {}
	for str in string.gmatch( str, "([^"..sep.."]+)" ) do
		t[#t+1] = str
        end
	return t
	end

local function present( k, list )	-- returns true if value is in list
	for i = 1,#list do if k == list[i] then return true end end
	return false
	end

local function loadData( filename, mode )	-- read text file
	if lfs.exists( filename ) then
		local tab, nodes = {}, {}			-- tab can be edges or calls
		for line in lfs.lines( filename ) do
			if mode == 'edges' then
				local spl = split( line, ',' )
				table.insert( tab, {origin = spl[1],target = spl[2]})
				if not present(spl[1],nodes) then table.insert(nodes,spl[1]) end
				if not present(spl[2],nodes) then table.insert(nodes,spl[2]) end
				elseif mode == 'calls' then
					local spl = split( line, ',' )
					table.insert( tab, { spl[1], spl[2] } )
				else
					table.insert( tab, line)
				end
			end
		return tab, nodes
		else print(filename..' not found!')
		end
	end
-- ------------------------------------------------
-- Module
-- ------------------------------------------------

local Graph = {}

function Graph.new( dir )
    local self		= {}
    local nodes		= {}	-- Contains all nodes in the graph
    local edges		= {}	-- Contains all edges in the graph
    local edgeIDs	= 0		-- Used to create a unique ID for new edges

    local function resetBoundaries()		-- graph boundaries
        minX, maxX, minY, maxY = nil, nil, nil, nil
		end

    local function updateBoundaries( nx, ny )	-- update area enclosing graph
        return math.min( minX or nx, nx ), math.max( maxX or nx, nx ), math.min( minY or ny, ny ), math.max( maxY or ny, ny )
		end

    function self:addEdge( origin, target )	-- create edge between two nodes
        for _, edge in pairs( edges ) do
            if edge.origin == origin and edge.target == target then
                error "Trying to connect nodes which are already connected."
				end
			end

		if origin ~= target then
			edges[edgeIDs] = { id = id, origin = origin, target = target }
			edgeIDs = edgeIDs + 1
			end
		end

    -- Adds a node to the graph.
    -- @param id - The ID will be used to reference the Node inside of the graph.
    -- @param x  - The x coordinate the Node should be spawned at (optional).
    -- @param y  - The y coordinate the Node should be spawned at (optional).
    -- @param anchor - Wether the node should be locked in place or not (optional).
    -- @param ... - Additional parameters (useful when a custom Node class is used).

    function self:addNode( id, x, y, anchor, ... )
        nodes[id] = Node.new( id, x, y, anchor, ... )
        return nodes[id]	-- spawning additional nodes.
		end

    function self:removeNode( node )	-- remove node and attached edges
        nodes[node:getID()] = nil
        self:removeEdges( node )
		end

    function self:removeEdges( node )	-- remove edges attached to a node
        for id, edge in pairs( edges ) do
            if edge.origin == node or edge.target == node then
                edges[id] = nil
				end
			end
		end

    function self:update( dt, nodeCallback, edgeCallback )	-- update graph
        for _, edge in pairs( edges ) do
            edge.origin:attractTo( edge.target )
            edge.target:attractTo( edge.origin )
            if edgeCallback then edgeCallback( edge ) end
			end

        resetBoundaries()

        for _, nodeA in pairs( nodes ) do
            if not nodeA:isAnchor() then
                for _, nodeB in pairs( nodes ) do
                    if nodeA ~= nodeB then
                        nodeA:repelFrom( nodeB )
						end
					end
                nodeA:move( dt )
				end

            if nodeCallback then
                nodeCallback( nodeA )
				end

            minX, maxX, minY, maxY = updateBoundaries( nodeA:getPosition() )
			end
		end

	 function self:draw(  )		-- drawing functions
		if show.e then
			for _, edge in pairs( edges ) do
				gr.setColor( color.edge )
				gr.line( edge.origin:getX(), edge.origin:getY(), edge.target:getX(), edge.target:getY() )
				gr.line( edge.origin:getX()-2, edge.origin:getY()-2, edge.target:getX(), edge.target:getY() )
				gr.line( edge.origin:getX()+2, edge.origin:getY()+2, edge.target:getX(), edge.target:getY() )
				end
			end
		if calls then
			for i = 1, #calls do
				local call = {
					nodes[calls[i][1]]:getX(), nodes[calls[i][1]]:getY(),
					nodes[calls[i][2]]:getX(), nodes[calls[i][2]]:getY()}
				if show.c and call then
					gr.setColor( color.calls )
					gr.line( call[1],call[2],call[3],call[4])
					end
				end
			end

		for _, node in pairs( nodes ) do
			local x, y = floor( node:getX() ), floor( node:getY() )
			if node:getID() == firstNode then gr.setColor(color.center) else gr.setColor( color.node ) end
			if show.p then
				gr.draw( sprite.image, x, y, 0, sprite.scale, sprite.scale, sprite.offset, sprite.offset )
				end
			if show.l then
				gr.setColor( color.text )
				gr.print( node:getID(), x , y , -rotate.angle ,1,1,-fontsize/2,fontsize/2)
				end
			end
		end

    function self:getNode( id ) return nodes[id] end

	function self:getNodeAt(x, y, range)	-- finds closest node to cursor
        for _, node in pairs( nodes ) do
            local nx, ny = node:getPosition()
            if x < nx + range and x > nx - range and y < ny + range and y > ny - range then
                return node
				end
			end
		end

    function self:getBoundaries() return minX, maxX, minY, maxY end

    function self:setAnchor( id, x, y )	-- node stationary, unmoveable by applied force
        nodes[id]:setPosition( x, y )
        nodes[id]:Anchor( true )
		end

	function self.showBoundingBox()
		gr.setColor(color.bounding)
		local mX, mY = minX - sprite.size/2, minY - sprite.size/2	-- keep nodes in box
		local wX, hY = maxX - minX + sprite.size, maxY - minY + sprite.size
		gr.rectangle( 'line', mX, mY, wX, hY )
		gr.circle( 'fill', wX/2+mX, hY/2+mY, 5 )	-- center of box
		end

	function self.getStats()
		gr.setColor( color.stats )
		local stats = gr.getStats()
		local x, y, s = 10, 12, 20		-- location of text 'x','y' and line separation 's'
		for i, j in pairs( stats ) do
			y = y + s
			gr.print( ( '%s:  %s' ):format( i, j ), x, y )
			end
		gr.print( 'fps: '..love.timer.getFPS(), x, y + s )
		end

	function self.save( filename )
		local screenshot = gr.newScreenshot()
		screenshot:encode( 'png', filename .. '.png' )
		end

	love.window.setTitle( dir )

	local edg, nod = loadData( dir..'/edges.txt', 'edges' )	-- loads interconnect text data

	calls = loadData( dir..'/calls.txt', 'calls' )			-- [loads function calls]

	for i = 1, #nod do		-- create nodes
        self:addNode( nod[i], love.math.random( 0, gr.getWidth() ), love.math.random( 0, gr.getHeight() ))
		end

	for i = 1, #edg do		-- create edges
		 self:addEdge( nodes[edg[i].origin], nodes[edg[i].target])
		end

	firstNode = nod[1]	-- anchor first node to center of screen
    self:setAnchor( firstNode, gr.getWidth() * 0.5, gr.getHeight() * 0.5 )

    return self
	end

return Graph
